from PyQt5.QtGui import QColor
from qgis.core import QgsAbstractGeometry,QgsCurve,QgsFeature,QgsGeometry,QgsGeometryCollection,QgsLineString,QgsLineSymbolLayer,QgsMarkerSymbolLayer,QgsPoint,QgsPolygon,QgsRenderContext,QgsSimpleFillSymbolLayer,QgsRuleBasedRenderer,QgsSymbol,QgsSymbolLayer,QgsVectorLayer,QgsWkbTypes
from qgis.gui import QgisInterface
from ..com.orbitgis.tool.message.Message import Message
from ..orbitgt.mapping3d.sdk.viewer.AList import AList
from ..orbitgt.mapping3d.sdk.viewer.Color import Color
from ..orbitgt.mapping3d.sdk.viewer.Coordinate import Coordinate
from ..orbitgt.mapping3d.sdk.viewer.CoordinatePart import CoordinatePart
from ..orbitgt.mapping3d.sdk.viewer.CoordinatePath import CoordinatePath
from ..orbitgt.mapping3d.sdk.viewer.MapObject import MapObject
from ..orbitgt.mapping3d.sdk.viewer.Viewer3DM import Viewer3DM
from .MapObjects import MapObjects

# Class Overlays manages the overlay of QGIS features on the 3D Mapping viewer
#
class Overlays:
    # define the (outline) color of the overlays
    COLOR: Color = Color.fromRGBA(0,0,255,255)
    # define the fill color of the overlays
    FILL_COLOR: Color = Color.fromRGBA(0,0,255,128)
    # define the outline line width
    LINE_WIDTH: float = 1.0

    # Convert a ring to a part
    #
    def ringToPart(crs: str,ring: QgsCurve,hasZ: bool) -> CoordinatePart:
        part = CoordinatePart()
        part.crs = crs
        for point in ring.points(): # QgsPoint
            if (hasZ): part.points.add(Coordinate.fromXYZ(crs,point.x(),point.y(),point.z()))
            else: part.points.add(Coordinate.fromXY(crs,point.x(),point.y()))
        return part
    
    # Create a new map object from a geometry
    #
    def createMapObject(crs: str,ageometry: QgsAbstractGeometry,layerName: str,objectId: str,mapObjects: AList[MapObject]) -> None:
        if (ageometry==None):
            return
        hasZ = ageometry.is3D()
        # do we have a point?
        if (isinstance(ageometry,QgsPoint)):
            point = ageometry
            Message.print(__name__,"Adding point '"+objectId+"'")
            mapObject = MapObject()
            if (hasZ): mapObject.spatialPoint = Coordinate.fromXYZ(crs,point.x(),point.y(),point.z())
            else: mapObject.spatialPoint = Coordinate.fromXY(crs,point.x(),point.y())
            mapObject.color = Overlays.COLOR
            mapObject.resourceId = layerName
            mapObject.objectId = objectId
            mapObjects.add(mapObject)
            return
        # do we have a polyline?
        if (isinstance(ageometry,QgsLineString)):
            lineString = ageometry
            Message.print(__name__,"Adding line-string '"+objectId+"'")
            linePart = CoordinatePart()
            linePart.crs = crs
            for point in lineString.points(): # QgsPoint
                if (hasZ): linePart.points.add(Coordinate.fromXYZ(crs,point.x(),point.y(),point.z()))
                else: linePart.points.add(Coordinate.fromXY(crs,point.x(),point.y()))
            mapObject = MapObject()
            mapObject.spatialLine = CoordinatePath()
            mapObject.spatialLine.crs = crs
            mapObject.spatialLine.parts.add(linePart)
            mapObject.color = Overlays.COLOR
            mapObject.lineWidth = Overlays.LINE_WIDTH
            mapObject.resourceId = layerName
            mapObject.objectId = objectId
            mapObjects.add(mapObject)
            return
        # do we have a polygon?
        if (isinstance(ageometry,QgsPolygon)):
            Message.print(__name__,"Adding polygon '"+objectId+"'")
            polygon = ageometry
            mapObject = MapObject()
            mapObject.spatialArea = CoordinatePath()
            mapObject.spatialArea.crs = crs
            mapObject.spatialArea.parts.add(Overlays.ringToPart(crs,polygon.exteriorRing(),hasZ))
            for i in range(0,polygon.numInteriorRings()): mapObject.spatialArea.parts.add(Overlays.ringToPart(crs,polygon.interiorRing(i),hasZ))
            mapObject.color = Overlays.COLOR
            mapObject.fillColor = Overlays.FILL_COLOR
            mapObject.lineWidth = Overlays.LINE_WIDTH
            mapObject.resourceId = layerName
            mapObject.objectId = objectId
            mapObjects.add(mapObject)
            return
        # do we have a geometry collection?
        if (isinstance(ageometry,QgsGeometryCollection)): # includes QgsMultiLineString and QgsMultiPolygon
            Message.print(__name__,"Found geometry collection")
            geometryCollection = ageometry
            Message.print(__name__,"Adding "+str(geometryCollection.childCount())+" child geometries")
            for i in range(0,geometryCollection.childCount()):
                Overlays.createMapObject(crs,geometryCollection.childGeometry(i),layerName,objectId+"["+str(i)+"]",mapObjects)
            return
        Message.print(__name__,"Feature not added as overlay, the geometry class is: "+ageometry.__class__.__name__)
        return
    
    # Create new map objects from a feature
    #
    def createMapObjects(crs: str,feature: QgsFeature,layerName: str,mapObjects: AList[MapObject]) -> None:
        geometry = feature.geometry() # QgsGeometry
        if (geometry==None):
            Message.print(__name__,"Feature has no geometry")
            return
        ageometry = geometry.constGet() # QgsAbstractGeometry
        if (ageometry==None):
            Message.print(__name__,"Feature has no abstract geometry")
            return
        objectId = layerName+"#"+str(feature.id())
        Overlays.createMapObject(crs,ageometry,layerName,objectId,mapObjects)
    
    # Convert a QGIS color to an SDK color
    #
    def getColor(color: QColor) -> Color:
        return Color.fromRGBA(color.red(),color.green(),color.blue(),color.alpha())

    # Update the legend of the map object
    #
    def updateLegend(layer: QgsVectorLayer,feature: QgsFeature,mapObject: MapObject,renderContext: QgsRenderContext):
        try:
            #Message.print(__name__,"Updating mapobject legend")
            symbol = None
            if (isinstance(layer.renderer(),QgsRuleBasedRenderer)): # the rule-based renderer does not implement the 'symbolForFeature' method well  (fix for the JIRA CLOUD-1076 issue)
                symbolList = layer.renderer().symbolsForFeature(feature,renderContext)
                if (len(symbolList)>0): symbol = symbolList[0]
            else:
                symbol = layer.renderer().symbolForFeature(feature,renderContext)
            if (symbol==None):
                return
            # check the first symbol layer to find the fill and outline colors
            symbolLayer = symbol.symbolLayer(0)
            if (isinstance(symbolLayer,QgsMarkerSymbolLayer)):
                #Message.print(__name__,"Found marker symbol layer")
                mapObject.color = Overlays.getColor(symbolLayer.strokeColor())
                mapObject.fillColor = Overlays.getColor(symbolLayer.color())
            if (isinstance(symbolLayer,QgsLineSymbolLayer)):
                #Message.print(__name__,"Found line symbol layer")
                #mapObject.lineWidth = symbolLayer.width()
                mapObject.color = Overlays.getColor(symbolLayer.color())
            if (isinstance(symbolLayer,QgsSimpleFillSymbolLayer)):
                #Message.print(__name__,"Found fill symbol layer")
                #mapObject.lineWidth = symbolLayer.strokeWidth()
                mapObject.color = Overlays.getColor(symbolLayer.strokeColor())
                mapObject.fillColor = Overlays.getColor(symbolLayer.fillColor())
        except Exception as e:
            Message.printError(__name__,e,"Error while setting legend of overlay object")
        
    # Add the selected QGIS features as overlays on the 3D Mapping viewer
    #
    def doAddOverlays(iface: QgisInterface,viewer: Viewer3DM) -> None:
        try:
            Message.print(__name__,"Adding the selected fatures as overlays")
            # get the current layer
            layer = iface.mapCanvas().currentLayer()
            layerName = layer.name()
            # we should have a vector layer
            if (isinstance(layer,QgsVectorLayer)==False):
                Message.print(__name__,"The current layer '"+layerName+"' is not a vector layer")
                return
            # get the epsg crs code of the layer
            layerCRS = MapObjects.getEpsgCode(layer.sourceCrs())
            if (layerCRS==None):
                Message.print(__name__,"The current vector layer '"+layerName+"' does not have an EPSG CRS")
                return
            # create a map object for each selected feature
            Message.print(__name__,"Found vector layer '"+layerName+"', crs '"+layerCRS+"', type "+QgsWkbTypes.displayString(layer.wkbType()))
            Message.print(__name__,"The layer renderer is: "+layer.renderer().__class__.__name__)
            mapObjects = AList[MapObject](MapObject.TYPE)
            renderContext = QgsRenderContext()
            layer.renderer().startRender(renderContext,layer.fields()) # put the renderer (QgsFeatureRenderer) in a state able to evaluate expressions (fix for the JIRA CLOUD-1077 issue)
            for feature in layer.getSelectedFeatures():
                featureMapObjects = AList[MapObject](MapObject.TYPE)
                Overlays.createMapObjects(layerCRS,feature,layerName,featureMapObjects)
                for i in range(0,featureMapObjects.size()):
                    mapObject = featureMapObjects.get(i)
                    Overlays.updateLegend(layer,feature,mapObject,renderContext)
                    mapObjects.add(mapObject)
            layer.renderer().stopRender(renderContext)
            Message.print(__name__,"Created "+str(mapObjects.size())+" map objects")
            # send the map objects to the viewer
            if (mapObjects.size()>0): viewer.addOverlayObjects(mapObjects)
            else: viewer.clearOverlayObjects()
            # clear the feature selection to indicate we have processed the objects
            layer.removeSelection()
        except Exception as e:
            Message.printError(__name__,e,"Error while adding overlays")
